(***********************************************************************)
(*                              Project Couverture                     *)
(*                                                                     *)
(* file: cover.ml                                                      *)
(* authors: Adrien Jonquet, Philippe Wang, Alexis Darrasse             *)
(* licence: CeCIL-B                                                    *)
(***********************************************************************)

let all_cover_res = ref ([] : Coverage.statement_result list) 
let cov_status = ref ([||] : Coverage.coverage_status array)
let events = ref ([||] : CoverageUtils.event list array)
let exec = ref ""
let exec_trace_list = ref ([] :  ((Trace.statement_trace array) list))
let exec_trace = ref ([||] : (Trace.statement_trace array))
let mcdc_trace_list = ref ([] :  ((string * (TraceMcdc.mcdc list)) list) list)
let mcdc_trace = ref ([] :  (string * (TraceMcdc.mcdc list)) list)
  
exception No_mcdc_info

let init_cover exe data exec_trace_files mcdc_trace_files =
  let code_section_size = Array.length data.Bytecode_loader.code_section in
    (* initialize the array of coverage events *)
    exec := exe;
    (* read the execution trace files *)
    List.iter
      (fun s ->
	 (* check if the trace exists *)
	 if Sys.file_exists s then
           begin
             (* open traces *)
             let ci = open_in s in
               exec_trace_list := (Marshal.from_channel ci) :: !exec_trace_list;
               close_in ci;
           end
	 else
           failwith ("exec trace file "^s^" does not exist"))
      exec_trace_files;
    (* init the execution trace resulting of the fusion of the execution trace files *)
    exec_trace := Array.make (Array.length (List.hd !exec_trace_list)) Trace.No_instr;
    (* read the mcdc trace files *)
    List.iter
      (fun s ->
	 (* check if the trace exists *)
	 if Sys.file_exists s then
           begin
             (* open mcdc traces *)
             let ci = open_in s in
               mcdc_trace_list := (Marshal.from_channel ci) :: !mcdc_trace_list;
               close_in ci;
           end
	 else
           failwith ("mcdc trace file "^s^" does not exist"))
      mcdc_trace_files;
    (* build the array of events *)
    events := Array.make code_section_size ([]:CoverageUtils.event list);
    CoverageUtils.get_module_names ();    
    List.iter 
      (fun module_name -> CoverageUtils.record_events !events data module_name)
      !CoverageUtils.module_names;
    (* build the array of bytecode instr for object code coverage *)
    cov_status := Array.make code_section_size Coverage.Uncovered;
    if not (Sys.file_exists "coverage_report/") then
      Unix.mkdir "coverage_report/" 0o755;
    if not (Sys.file_exists "coverage_report/coverage.css") then
      HtmlReport.generate_css (open_out "coverage_report/coverage.css")
	
let cover_object data =
  (* check which instruction has been executed *)
  Coverage.cover_object !cov_status !exec_trace_list;
  (* build the html report *)
  HtmlReport.generate_html_asm !exec data !cov_status

(* fusion of the execution trace files *)
let fusion_exec_trace exec_trace_files = 
  List.iter (fun trace_file ->
	       for i=0 to (Array.length trace_file)-1 do
		 match trace_file.(i) with
		   | Trace.No_instr -> ()
		   | Trace.Trace_hist t as x -> !exec_trace.(i) <- x
	       done
	    ) exec_trace_files
    
(* fusion the vectors of the mcdc trace files *)
let fusion_mcdc mcdc_trace_files =
  let vector_fusion vectors_res vectors =
    let module Fusion = 
      Set.Make (struct type t=TraceMcdc.vector 
		       let compare = Pervasives.compare
		end) 
    in
      Fusion.elements  
	(List.fold_left (fun r e -> Fusion.add e r) 
	   (List.fold_left (fun r e -> Fusion.add e r) Fusion.empty vectors_res)
	   vectors)
  in
  let rec add_to_res mod_name mcdc_list = function
      (* module never encountered -> put directly in the final mcdc res *)
      [] -> mcdc_trace := (mod_name, mcdc_list) :: !mcdc_trace
    | (mod_name_res, mcdc_list_res) :: l ->
	if mod_name = mod_name_res then
	  (* module encountered -> vectors fusion *)
	  List.iter 
	    (fun mcdc_res -> 
	       List.iter 
		 (fun mcdc ->
		    (* same boolean expression *)
		    if mcdc.TraceMcdc.model = mcdc_res.TraceMcdc.model then
		      mcdc_res.TraceMcdc.vectors <- 
			vector_fusion mcdc.TraceMcdc.vectors mcdc_res.TraceMcdc.vectors
		 ) mcdc_list
	    ) mcdc_list_res
	else
	  add_to_res mod_name mcdc_list l
  in
    List.iter 
      (fun file ->
	 List.iter 
	   (fun (module_name, mcdc_list) ->
	      add_to_res module_name mcdc_list !mcdc_trace) file) 
      mcdc_trace_files

(* expression coverage *)
(* MC/DC coverage *)
let cover_source exec_trace_files mcdc_trace_files =
  (* mcdc_trace fusion *)
  fusion_mcdc !mcdc_trace_list;
  fusion_exec_trace !exec_trace_list;
  List.iter
    (fun src ->
       (* for each .ml source file *)
       if Sys.file_exists src then
	 begin
	   (* Expression coverage *)
	   let expression_res = Coverage.cover_expression src !exec_trace !events in
	   let mcdc_res = Coverage.cover_mcdc src !mcdc_trace in
	   let cover_res = {
             Coverage.src = expression_res.Coverage.src;
             Coverage.statements = 
	       mcdc_res.Coverage.statements @ expression_res.Coverage.statements;
             Coverage.nb_events = 
	       mcdc_res.Coverage.nb_events + expression_res.Coverage.nb_events;
             Coverage.nb_uncovered_events = 
	       mcdc_res.Coverage.nb_uncovered_events 
	       + expression_res.Coverage.nb_uncovered_events
	   }
	   in
	     all_cover_res := expression_res :: !all_cover_res;
	     let rec aux = function
	       | (module_name, mcdc :: l2) :: l1 ->
	           begin 
    	             let module_name =  
		       String.capitalize (String.sub src 0 ((String.length src) - 3)) in
		       if mcdc.TraceMcdc.model.TraceMcdc.mod_dec.Covevent.covev_module 
			 = module_name then (module_name, mcdc :: l2)
		       else aux l1
	           end
	       | _ -> raise No_mcdc_info
	     in
	       (* HTML report generation *)
	       begin
		 try
		   HtmlReport.generate_mcdc_tabular src (aux (List.hd !mcdc_trace_list)) cover_res;
		 with
		   | No_mcdc_info -> print_string (src ^ " ---> No boolean expression\n")
	       end;
	       (* HTML report generation *)
	       HtmlReport.generate_html src cover_res 
	 end)
    !CoverageUtils.src_list;
  HtmlReport.generate_html_index !exec !all_cover_res exec_trace_files mcdc_trace_files
    
    
    
    
