(***********************************************************************)
(*                              Project Couverture                     *)
(*                                                                     *)
(* file: coverage.ml                                                   *)
(* authors: Adrien Jonquet, Philippe Wang, Alexis Darrasse             *)
(* licence: CeCIL-B                                                    *)
(***********************************************************************)

open CoverageUtils

type coverage_status = Uncovered | Covered 

type coverage_expr = {
  event : EventType.event;
  cnum_start : int;
  cnum_end : int;
  mutable status : coverage_status;
}

type statement_result = {
  src : string;
  statements : coverage_expr list; 
  nb_events : int;
  nb_uncovered_events : int
}

(* Verify wich bytecode instruction has been executed *)
let cover_object cov_status trace_list =
  let rec is_executed pc l = 
    match l with
      | [] -> false
      | trace :: trace_list -> 
	  match trace.(pc) with
	    | Trace.No_instr -> is_executed pc trace_list
	    | Trace.Trace_hist t -> true
  in
    for i = 0 to (Array.length cov_status) - 1 do
      if is_executed i trace_list then
	cov_status.(i) <- Covered
    done    

let cover_expression src exec_trace events =
  let statements = ref [] in
  let nb_uncovered = ref 0 in
  let nb_ev = ref 0 in
  let is_executed pc exec_trace =
    match exec_trace.(pc) with
      | Trace.No_instr -> false
      | Trace.Trace_hist t -> true
  in
  let rec glance_events i = function
      [] -> ()
    | ev :: l ->
	(* Look only for events associated with the src file *)
	begin
	  if (get_pos ev).Lexing.pos_fname = src then
	    begin
	      incr nb_ev;
	      let l, l' = get_loc ev in
		match ev.Covevent.covev_status with
		  | Covevent.Coverage_event_condition _ -> ()
		  | Covevent.Coverage_event_decision -> ()
		  | Covevent.Coverage_event_statement ->
		      if is_executed i exec_trace then
			begin
			  statements :=
			    { event = ev;
			      cnum_start = l.Lexing.pos_cnum;
			      cnum_end = l'.Lexing.pos_cnum;
			      status = Covered } :: !statements;
			end
		      else
			begin
			  statements :=
			    { event = ev;
			      cnum_start = l.Lexing.pos_cnum;
			      cnum_end = l'.Lexing.pos_cnum;
			      status = Uncovered } :: !statements;
			  incr nb_uncovered
			end 
	    end;
	  glance_events i l
	end
  in
    for i=0 to (Array.length events) - 1 do
      glance_events i events.(i)
    done;
    {
      src = src;
      statements = !statements;
      nb_events = !nb_ev;
      nb_uncovered_events = !nb_uncovered
    }
      
(* analyse the MC/DC vectors *)
let is_independant cond vectors = 
  let is_mcdc_complementary vec_conds1 vec_conds2 =
    let res = ref true in   
      for i = 0 to (Array.length vec_conds1) - 1 do
        if (vec_conds1.(i).TraceMcdc.cond_value <> TraceMcdc.Unknown) 
          && (vec_conds2.(i).TraceMcdc.cond_value <> TraceMcdc.Unknown) then
            if vec_conds1.(i).TraceMcdc.cond.Covevent.covev_loc = cond.Covevent.covev_loc then
              begin
                if vec_conds1.(i).TraceMcdc.cond_value = vec_conds2.(i).TraceMcdc.cond_value then
                  res := false
              end
            else
              begin
                if (vec_conds1.(i).TraceMcdc.cond_value <> TraceMcdc.Unknown) 
                  && (vec_conds2.(i).TraceMcdc.cond_value <> TraceMcdc.Unknown) then
                    if vec_conds1.(i).TraceMcdc.cond_value <> vec_conds2.(i).TraceMcdc.cond_value then
                      res := false
              end
      done;
      !res
  in
  let rec check_independance vector l = 
    let b = ref true in
      begin
	match l with
            [] -> b:= false
          | t :: q when (is_mcdc_complementary t.TraceMcdc.vec_conds vector.TraceMcdc.vec_conds) -> 
              begin
                if t.TraceMcdc.vec_dec.TraceMcdc.dec_value <> vector.TraceMcdc.vec_dec.TraceMcdc.dec_value then
                  b := true
                else b:= check_independance vector q
              end
          | t :: q -> b:= check_independance vector q
      end;
      !b
  in
  let rec glance_vectors vectors = 
    match vectors with
        [] -> false
      | vect :: l  -> 
          if check_independance vect vectors then true
          else glance_vectors l
            
  in 
    glance_vectors vectors
      
let is_mcdc mcdc =
  let b = ref true in
    for i = 0 to (List.length mcdc.TraceMcdc.model.TraceMcdc.mod_conds)-1 do
      if not (is_independant (List.nth mcdc.TraceMcdc.model.TraceMcdc.mod_conds i) mcdc.TraceMcdc.vectors) then
        b := false
    done;
    !b
      
let cover_conditions conditions mcdc =
  for i = 0 to (List.length mcdc.TraceMcdc.model.TraceMcdc.mod_conds)-1 do
    if not (is_independant (List.nth mcdc.TraceMcdc.model.TraceMcdc.mod_conds i) mcdc.TraceMcdc.vectors) then
      (List.nth conditions i).status <- Uncovered 
  done
    
let cover_mcdc src mcdc_trace =
  let res = ref [] in     
  let check_mcdc mcdc =
    let decision =
      begin 
        let l, l' = 
          get_loc mcdc.TraceMcdc.model.TraceMcdc.mod_dec in
          { event = mcdc.TraceMcdc.model.TraceMcdc.mod_dec;
            cnum_start = l.Lexing.pos_cnum;
	    cnum_end = l'.Lexing.pos_cnum;
	    status = Uncovered }
      end
    in
      (* Trivial Case: covering a decisions with not enough vectors to be covered *)
      if (List.length mcdc.TraceMcdc.vectors) <= (List.length mcdc.TraceMcdc.model.TraceMcdc.mod_conds) then
        decision.status <- Uncovered
      else
        begin
          (* Easy case: covering a simple decision with only one condition *)
	  (*            checks if the condition took both values (true and false) *)
	  if mcdc.TraceMcdc.model.TraceMcdc.mod_conds = [] then
	    begin
	      let dec_false = ref false in
	      let dec_true = ref false in
	      let rec check_dec = function
		  [] -> ()
		| vect :: l -> 
		    begin 
		      match vect.TraceMcdc.vec_dec.TraceMcdc.dec_value with
			  TraceMcdc.Bool true 
			    when vect.TraceMcdc.vec_dec.TraceMcdc.nb_dec > 0 -> 
			      dec_true := true
			| TraceMcdc.Bool false 
			    when vect.TraceMcdc.vec_dec.TraceMcdc.nb_dec > 0 -> 
			    dec_false := true
			| _ -> ()
		    end;
		    check_dec l
	      in
		check_dec  mcdc.TraceMcdc.vectors;
		if !dec_false && !dec_true then
		  decision.status <- Covered
		else 		  
		  decision.status <- Uncovered;
		 dec_true := false;
		 dec_false := false;
	    end
          else 
            begin 
              if is_mcdc mcdc then 
                decision.status <- Covered
            end
        end;
      let conditions = 
	let tmp = ref [] in
          List.iter (fun cond ->
		       let l, l' = get_loc cond in
			 tmp := { event = cond;
			          cnum_start = l.Lexing.pos_cnum;
			          cnum_end = l'.Lexing.pos_cnum;
			          status = Covered } :: !tmp)
            mcdc.TraceMcdc.model.TraceMcdc.mod_conds;
          !tmp
      in
	cover_conditions (List.rev conditions) mcdc;
	res := (decision :: conditions) @ !res;
  in
   List.iter
      (fun (module_name, mcdc_list) -> 
	 List.iter 
	   (fun mcdc ->
	      if (get_pos mcdc.TraceMcdc.model.TraceMcdc.mod_dec).Lexing.pos_fname = src 
	      then check_mcdc mcdc) 
	   mcdc_list)
      mcdc_trace;
    { src = src;
      statements = !res;
      nb_events = 0;
      nb_uncovered_events = 0 }
      
