(***********************************************************************)
(*                              Project Couverture                     *)
(*                                                                     *)
(* file: htmlReport.ml                                                 *)
(* authors: Adrien Jonquet, Philippe Wang, Alexis Darrasse             *)
(* licence: CeCIL-B                                                    *)
(***********************************************************************)

open HtmlReportUtils

let head co title =
  output_string co
    ("<HEAD>
      <TITLE>"^title^"</TITLE>
      <LINK rel=\"stylesheet\" type=\"text/css\" href=\"coverage.css\">
      <META http-equiv='Content-Type' content='text/html; charset=ISO-8859-1'>
      <SCRIPT TYPE=text/javascript>
        function show(x) {
          x.style.fontStyle = 'italic';
          x.style.fontWeight = 'bold';
        }
        function hide(x) {
          x.style.fontStyle = 'normal';
          x.style.fontWeight = 'normal';
        }
      </SCRIPT>
  </HEAD>") 
    
let head_object co =
  output_string co
    "<HEAD>
    <TITLE>Object code coverage report</TITLE>
      <LINK rel=\"stylesheet\" type=\"text/css\" href=\"coverage.css\">
    <META http-equiv='Content-Type' content='text/html; charset=ISO-8859-1'>
  </HEAD>" 

(*.covered_decision { margin:1px;padding:1px;border:dashed;border-color:blue; }
  .uncovered_decision { margin:1px;padding:1px;border:dashed;border-color:red; }*)

let generate_css co =
  output_string co 
    "
.bytecode  {font-family:monospace}

.covered { background-color: #55FF55 }
.uncovered { background-color: #FF6633 }

.covered_decision { border-bottom: 2px solid #55FF55; }
.uncovered_decision { border-bottom: 2px solid #FF6633; }
.covered_condition { background-color: #55FF55 }
.uncovered_condition { background-color: Fuchsia }

tr.IndexSummary td.indexSummary {
  background-color: #B0C4DE;
  border-color: white;
  border-style: double;
}
td.SumBarUncovered {
  background-color: Red;
}
td.SumBarCovered {
  background-color: Lime;
}
"
let draw_visual_bar co percentage =
  let percentage_uncov = 100 - percentage in
    output_string co ("<table border=\"0\" cellspacing=\"0\">");
    output_string co ("<tr height=\"10\">");
    if percentage <> 0 then
      output_string co ("<td class=SumBarCovered width=\""^(string_of_int percentage)^"\"></td>");
    if percentage_uncov <> 0 then
      output_string co ("<td class=SumBarUncovered width=\""^(string_of_int percentage_uncov)^"\"></td>");
    output_string co ("</tr></table>")

let generate_html src cover_res =
  try  
    begin
      if Sys.file_exists src then
	begin
	  let ci = open_in src in
	  let co = open_out ("coverage_report/"^src^".html") in
	    output_string co "<HTML>";
	    head co "Coverage report";
	    (* write BODY *)
	    output_string co ("<BODY><h3>ZamCov: Expression Coverage <br>("^src^")</h3><pre>\n");
	    output_string co ("<h2>MC/DC vectors : <a href=\""^src^".mcdc.html\">"^src^".mcdc</a></h2>\n\n");
	    let i = ref 0 in
	      begin 
		try
		  while true do
		    color_expression co !i cover_res.Coverage.statements;
		    let c = input_char ci in
		      output_char co c;
		      incr i;
		  done
		with
		  | End_of_file -> output_string co "</pre></BODY>"
	      end;
	      output_string co "</HTML>";
	      close_in ci;
	      close_out co
	end
    end
  with
    | Sys_error _ -> failwith "source file error html"
	
let generate_mcdc_tabular src (module_name, mcdc_trace) cover_res =
  try
    begin
      if Sys.file_exists src then
	begin
	  let ci = open_in src in
	  let co = open_out ("coverage_report/"^src^".mcdc.html") in
	    output_string co "<HTML>";
	    head co "MC/DC report";
	    (* write BODY *)
	    output_string co ("<BODY><h3>ZamCov: MC/DC tabulars <br>("^src^")</h3><pre>\n");
	    List.iter (fun mcdc ->
			 output_string co "<TABLE BORDER=\"1\">"; 	      
			 output_string co "<TR>";
			 output_string co "<TH>";	       
			 output_string co " DEC ";
			 List.iter (
			   fun statement ->
			     if statement.Coverage.event = mcdc.TraceMcdc.model.TraceMcdc.mod_dec then
	                       begin
				 match statement.Coverage.status with
	                             Coverage.Covered ->  output_string  co "<SPAN CLASS=covered>"
	                           |  Coverage.Uncovered -> output_string  co "<SPAN CLASS=uncovered>"
	                       end;
			 ) cover_res.Coverage.statements;
			 let l, l' = CoverageUtils.get_loc mcdc.TraceMcdc.model.TraceMcdc.mod_dec in
			 let line_start = l.Lexing.pos_lnum in
			   output_string co ("line "^(string_of_int line_start)^": ");
			   let cnum_start = l.Lexing.pos_cnum in
			   let cnum_end = l'.Lexing.pos_cnum in
			   let expr = String.create (cnum_end - cnum_start) in
			     seek_in ci cnum_start;
			     really_input ci expr 0 (cnum_end - cnum_start);
			     output_string co expr; 
			     output_string co "</SPAN>";
			     output_string co "</TH>";
			     
			     List.iter (fun cond ->
					  let l, l' = CoverageUtils.get_loc cond in
					  let cnum_start = l.Lexing.pos_cnum in
					  let cnum_end = l'.Lexing.pos_cnum in
					  let expr = String.create (cnum_end - cnum_start) in
					    seek_in ci cnum_start;
					    really_input ci expr 0 (cnum_end - cnum_start);
					    output_string co "<TH>";
					    output_string co expr; 
					    output_string co "</TH>"; 	
				       ) mcdc.TraceMcdc.model.TraceMcdc.mod_conds;
			     
			     List.iter (fun vector ->
					  output_string co "<TR>"; 	
					  output_string co "<TH>";
					  begin
					    match  vector.TraceMcdc.vec_dec.TraceMcdc.dec_value with 
						TraceMcdc.Bool true -> output_string co " T "
					      | TraceMcdc.Bool false -> output_string co " F "
					      | TraceMcdc.Unknown ->  output_string co " - "
					  end;
					  let nb = string_of_int vector.TraceMcdc.vec_dec.TraceMcdc.nb_dec in
					    output_string co (": "^nb);
					    output_string co "</TH>";

					    List.iter (fun cond ->
	               					 output_string co "<TH>";
	               					 for i=0 to (Array.length vector.TraceMcdc.vec_conds - 1) do
							   if cond = vector.TraceMcdc.vec_conds.(i).TraceMcdc.cond then
							     begin
							       match  vector.TraceMcdc.vec_conds.(i).TraceMcdc.cond_value with 
								   TraceMcdc.Bool true -> output_string co " T "
								 | TraceMcdc.Bool false -> output_string co " F "
								 | TraceMcdc.Unknown ->  output_string co " - "
							     end;
							 done;
							 output_string co "</TH>";
						      ) mcdc.TraceMcdc.model.TraceMcdc.mod_conds;
					    
					    output_string co "</TR>"; 	
				       ) mcdc.TraceMcdc.vectors;
			     
			     output_string co "</TR"; 	      
			     output_string co "</TABLE>";
     			     output_string co "<BR>";
		      ) mcdc_trace;
	    
	    output_string co "</HTML>";
	    close_in ci;
	    close_out co
	end
    end
  with
    | Sys_error _ -> failwith "source file error html"

let generate_html_index exec all_cover_res exec_trace_file_names mcdc_trace_file_names =
  let co = open_out ("coverage_report/index.html") in

    head co ("Code coverage report for :"^exec);
    output_string co "<BODY>\n";  
    output_string co "<h2>Coverage report</h2>\n";

    (***************************************************)
    (* write the list of trace files used for coverage *)
    (***************************************************)
    output_string co "<hr/>\n";
    output_string co "<table CELLPADDING=1 class=\"IndexSummary\">";
    
    output_string co "
<tr class=\"IndexSummary\">
  <td class=\"IndexSummary\" align=center><b>Program Name</b></td>
</tr>\n";
    output_string co "<tr class=\"IndexSummary\">\n";
    output_string co ("<td class=\"IndexSummary\">"^exec^"</td></tr>\n");
    output_string co "</table>";

    output_string co "<table CELLPADDING=1 class=\"IndexSummary\">";
 output_string co "
<tr class=\"IndexSummary\">
  <td class=\"IndexSummary\" align=center><b>Trace Filename</b></td>
</tr>\n";
    List.iter 
      (fun trace ->
	 output_string co "<tr class=\"IndexSummary\">\n";
	 output_string co ("<td class=\"IndexSummary\">"^trace^"</td></tr>\n"))
      exec_trace_file_names;
    output_string co "</table>";

    output_string co "<table CELLPADDING=1 class=\"IndexSummary\">";
 output_string co "
<tr class=\"IndexSummary\">
  <td class=\"IndexSummary\" align=center><b>MC/DC trace Filename</b></td>
</tr>\n";
    List.iter 
      (fun trace ->
	 output_string co "<tr class=\"IndexSummary\">\n";
	 output_string co ("<td class=\"IndexSummary\">"^trace^"</td></tr>\n"))
      mcdc_trace_file_names;
    output_string co "</table>";
    
    (******************************************)
    (* write the list of covered source files *)
    (******************************************)
    output_string co "<hr/>\n";
    output_string co "<table CELLPADDING=1 class=\"IndexSummary\">";
    output_string co "
<tr class=\"IndexSummary\">
  <td class=\"IndexSummary\" align=center>Source Filename</td>
  <td class=\"IndexSummary\" align=center>Expression coverage</td>
</tr>
";
    List.iter 
      (fun x ->
	 let src = x.Coverage.src in
	 let nb_events = x.Coverage.nb_events in
	 let nb_covered = nb_events - x.Coverage.nb_uncovered_events in
	 let percentage = int_of_float ((float nb_covered) /. (float nb_events) *. 100.) in
	   if Sys.file_exists src then 
	     begin
	       output_string co "<tr class=\"IndexSummary\"><td class=\"IndexSummary\">\n";
	       output_string co ("<a href=\""^src^".html\">\n");
	       output_string co (src^"</a></td>\n");
	       output_string co ("<td class=\"IndexSummary\" align=right>"^(string_of_int percentage)^"%  ");
	       output_string co ((string_of_int nb_covered)^"/ "^(string_of_int nb_events));
	       draw_visual_bar co percentage;
	     end)
      all_cover_res;
    output_string co "</table>\n";
    
    output_string co "<h2>Object code coverage report</h2>\n";
    output_string co ("<a href="^exec^".asm.html>"^exec^"</a>\n");
    output_string co "</BODY>";
    close_out co

let generate_html_asm exec data cov_status =
  let code = data.Bytecode_loader.code_section in
    try 
      let co = open_out ("coverage_report/"^exec^".asm.html") in
	output_string co "<html>";
	head_object co;
	output_string co "<body><div class='bytecode'>";
	let aux cl inst i =
          output_string co (Printf.sprintf "%08d <span class='%s'>" i cl);
	  output_string co (Instructions.string_of_instructions data.Bytecode_loader.primitive_section inst);
	  output_string co "</span><br/>";
	in
	  for i=0 to Array.length code - 1 do
            match code.(i),cov_status.(i) with
              | Instructions.Param _, _ -> ()
              | inst, Coverage.Uncovered -> aux "uncovered" inst i
              | inst, Coverage.Covered -> aux "covered" inst i
	  done;
	  output_string co "</div>";
	  output_string co "</body>";
	  output_string co "</html>";
	  close_out co
    with
      | Sys_error _ -> failwith "source file error obj"
