(***********************************************************************)
(*                              Project Couverture                     *)
(*                                                                     *)
(* file: traceMcdc.ml                                                  *)
(* authors: Adrien Jonquet, Mathias Bourgoin                           *)
(* licence: CeCIL-B                                                    *)
(***********************************************************************)

open EventType
open Instructions

let prims = ref (Array.make 0 "")
  
(* MC/DC vector type *)
type mcdc_bool = Bool of bool | Unknown

type condition = 
    { mutable cond_value : mcdc_bool;
      mutable nb_cond : int;
      mutable cond : event }
      
type decision = 
    { mutable dec_value : mcdc_bool;
      mutable nb_dec : int;
      mutable dec : event }
      
type vector_model =
    {   mutable mod_dec : event;
        mutable mod_conds : event list }

type vector =
    { mutable nb_vec : int;
      mutable vec_dec : decision;
      mutable vec_conds :  condition array}
      
type mcdc =
    { mutable model : vector_model;
      mutable vectors : vector list }

let last_cond = ref ([] : Covevent.coverage_event list)
let last_dec = ref ([] : Covevent.coverage_event list)

let nb_vectors = ref 0
let nb_merged = ref 0

let vector_completed vector = (vector.vec_dec.nb_dec > 0)


let rec merge_vector m vector = function
    [] -> (m.vectors <- List.tl m.vectors@[vector];)
  | v :: q -> 
      begin
	let b = ref true in
	  for i = 0 to (Array.length v.vec_conds)-1 do
	    if vector.vec_conds.(i).cond <> v.vec_conds.(i).cond then assert false;
	    (match vector.vec_conds.(i).cond_value, v.vec_conds.(i).cond_value with
	       | Unknown, Unknown -> ()
	       | Bool true, Bool true -> ()
	       | Bool false, Bool false -> ()
	       | _ -> b := false);
	  done;
	  if (!b) && (vector.vec_dec.dec_value = v.vec_dec.dec_value) then 
	    begin
	      v.vec_dec.nb_dec <- v.vec_dec.nb_dec + 1;
	      m.vectors <- List.tl m.vectors;
	    end
	  else
	    merge_vector m vector q 
      end

let rec record_dec dec v = function
    [] -> ()
  | m :: l ->
      if m.model.mod_dec = dec then
	begin
	  if m.model.mod_conds = [] then
	    begin
	      if m.vectors = [] then
		begin
		  m.vectors <- [ { nb_vec = 0;
				   vec_dec = { dec_value = Bool true; 
					       nb_dec = 0; 
					       dec = m.model.mod_dec };
				   vec_conds = [||]};
				 { nb_vec = 0; 
				   vec_dec = { dec_value = Bool false; 
					       nb_dec = 0; 
					       dec = m.model.mod_dec };
				   vec_conds = [||]} ];
		end;
	      if v then 
		begin
		  (List.nth m.vectors 0).nb_vec <- 
		    (List.nth m.vectors 0).nb_vec+1;
		  (List.nth m.vectors 0).vec_dec.nb_dec <-  
		    (List.nth m.vectors 0).vec_dec.nb_dec+1
		end
	      else
		begin
		  (List.nth m.vectors 1).nb_vec <- 
		    (List.nth m.vectors 1).nb_vec+1;
		  (List.nth m.vectors 1).vec_dec.nb_dec <-  
		    (List.nth m.vectors 1).vec_dec.nb_dec+1
		end;
	    end
	  else
            begin
              let vector = List.hd m.vectors in 
		vector.vec_dec <-  {dec_value = Bool v; nb_dec = 1; dec = m.model.mod_dec };
		last_dec := List.tl !last_dec;
	    end;
	  if (List.length m.vectors) > 1 then
	    merge_vector m (List.hd m.vectors) (List.tl m.vectors);
	end
      else
	record_dec dec v l     
	  

let rec record_cond cond v = function
    [] -> (last_cond := List.tl !last_cond)
  | m :: l ->
      if List.mem cond m.model.mod_conds then
	begin 
	  if (m.vectors == []) ||  (cond = (List.hd m.model.mod_conds) ) then
            begin
              let new_vec_cond = 
		Array.init (List.length m.model.mod_conds) 
		  (fun i -> 
		     if ((List.nth m.model.mod_conds i).Covevent.covev_loc = cond.Covevent.covev_loc) then
                       { cond_value = Bool v; 
			 nb_cond = 1; 
			 cond = (List.nth m.model.mod_conds i) }
                     else
                       { cond_value = Unknown; 
			 nb_cond = 0; 
			 cond = (List.nth m.model.mod_conds i) })
	      in
		m.vectors <- {nb_vec = 0; 
			      vec_dec = { dec_value= Unknown; 
					  nb_dec =0; 
					  dec = m.model.mod_dec };
			      vec_conds = new_vec_cond } :: m.vectors;
            end
          else
            begin
	      let vector = List.hd m.vectors in
		for i=0 to (Array.length vector.vec_conds)-1 do
		  if vector.vec_conds.(i).cond.Covevent.covev_loc = cond.Covevent.covev_loc then
		    begin
		      vector.vec_conds.(i).nb_cond <- vector.vec_conds.(i).nb_cond + 1;
		      vector.vec_conds.(i).cond_value <- Bool v
		    end
		done;
	    end;
        end;
      record_cond cond v l

let last_instr = ref ATOM0

(* record the boolean vectors for MC/DC *)
let record_mcdc vm mcdc_trace cond_list dec_list instr prims = 
  let rec find_cov_from_adr pc = function
      [] -> false
    | cov :: l -> 
        if cov.Covevent.covev_pos = pc then
          begin 
            match cov.Covevent.covev_status with
		Covevent.Coverage_event_condition _ -> 
		  last_cond := cov :: [](*!last_cond*);
                  true;
              | Covevent.Coverage_event_decision -> 
		  last_dec := cov :: [](*!last_dec*);
                  true; 
              | _ ->  false
          end
        else find_cov_from_adr pc l
  in 
    if find_cov_from_adr (vm.Vm.code_pointer * 4) (cond_list) then
      begin
	if !last_cond <> []  then
	  begin
	    match instr with
	      | BOOLNOT -> 
		  let v = vm.Vm.accumulator <> Value.Int 0 in
	            record_cond (List.hd !last_cond) v mcdc_trace
	      | BRANCHIF n -> 
	          let v = vm.Vm.accumulator <> Value.Int 0 in
	            record_cond (List.hd !last_cond) v mcdc_trace
	      | BRANCHIFNOT n ->
	          let v = vm.Vm.accumulator <> Value.Int 0 in
	            record_cond (List.hd !last_cond) v mcdc_trace
	      | RETURN n ->
	          (match !last_instr with
	               BRANCHIF n -> ()
	             | BRANCHIFNOT n -> () 
	             | _ -> let v = vm.Vm.accumulator <> Value.Int 0 in
			 record_cond (List.hd !last_cond) v  mcdc_trace
	          );
	      | ACC0 -> let v = Vm.peek vm 0 <> Value.Int 0 in
		  record_cond (List.hd !last_cond) v  mcdc_trace
	      | ACC1 -> let v = Vm.peek vm 1 <> Value.Int 0 in
		  record_cond (List.hd !last_cond) v  mcdc_trace
	      | ACC2 -> let v = Vm.peek vm 2 <> Value.Int 0 in
		  record_cond (List.hd !last_cond) v  mcdc_trace
	      | ACC3 -> let v = Vm.peek vm 3 <> Value.Int 0 in
		  record_cond (List.hd !last_cond) v  mcdc_trace
	      | ACC4 -> let v = Vm.peek vm 4 <> Value.Int 0 in
		  record_cond (List.hd !last_cond) v  mcdc_trace
	      | ACC5 -> let v = Vm.peek vm 5 <> Value.Int 0 in
		  record_cond (List.hd !last_cond) v  mcdc_trace
	      | ACC6 -> let v = Vm.peek vm 6 <> Value.Int 0 in
		  record_cond (List.hd !last_cond) v  mcdc_trace
	      | ACC7 -> let v = Vm.peek vm 7 <> Value.Int 0 in
		  record_cond (List.hd !last_cond) v  mcdc_trace
	      | ACC n -> let v = Vm.peek vm n <> Value.Int 0 in
		  record_cond (List.hd !last_cond) v  mcdc_trace
              | CONST0 -> record_cond (List.hd !last_cond) false  mcdc_trace
              | CONST1 -> record_cond (List.hd !last_cond) true  mcdc_trace
              |_ -> begin 
                        ( let v = vm.Vm.accumulator <> Value.Int 0 in
			 record_cond (List.hd !last_cond) v  mcdc_trace);  
                         print_string ("---"^(Instructions.string_of_instructions prims instr)); print_newline (); flush stdout;
                         print_string ("+++"^(Instructions.string_of_instructions prims !last_instr)); print_newline (); flush stdout;
                         assert false; exit 132;
                         end
	  end;
      end;
    if find_cov_from_adr (vm.Vm.code_pointer * 4) (dec_list) then
      begin
	if !last_dec <> [] then
          begin
	    let v = vm.Vm.accumulator <> Value.Int 0 in
	      record_dec (List.hd !last_dec) v mcdc_trace;
          end;
      end;
    last_instr := instr
      
(* build the boolean vectors for MC/DC *)
let build_mcdc_vectors data module_name =
  (* get the list of modules *)
  let ci = open_in (module_name ^ ".cmo.cover") in
  let (orig, ev_list) = Marshal.from_channel ci in
    close_in ci;
    if ev_list = [] then failwith "no coverage_events";
    List.iter (fun event ->
		 match event.Covevent.covev_status with
		     Covevent.Coverage_event_condition dec_list  ->
		       begin
			 List.iter (fun event -> CoverageUtils.relocate_event orig event);
			 CoverageUtils.relocate_event orig event
		       end
		   | _ ->  CoverageUtils.relocate_event orig event) 
      ev_list;
    (* get the list of decision and the list of condition *)        
    let rec first_pass l mcdc_l cond_l dec_l =
      match l with 
          [] -> mcdc_l, cond_l, dec_l
	| cov :: q -> 
            begin 
              match cov.Covevent.covev_status with
		| Covevent.Coverage_event_decision -> 
                    let vect = {
                      model = {mod_dec = cov; mod_conds = []};
                      vectors = []
                    }
                    in
                      first_pass q (vect::mcdc_l) cond_l (cov::dec_l)
		| Covevent.Coverage_event_condition _ -> first_pass q mcdc_l (cov::cond_l) dec_l
		| _ -> first_pass q mcdc_l cond_l dec_l
            end
    in
    let mcdc_list, cond_list, dec_list = 
      first_pass 
	(List.fold_left 
	   (fun lu event -> 
	      if List.exists 
		(fun a -> 
		   ((fst (CoverageUtils.get_loc a) = fst (CoverageUtils.get_loc event))  
		    && (snd (CoverageUtils.get_loc a) = snd (CoverageUtils.get_loc event)))) lu
	      then lu else (event :: lu))
	   [] ev_list) [] [] []
    in
    let rec populate_vector_model cond dec = function
	[] -> ()
      | t :: l when ((t.model).mod_dec.Covevent.covev_loc = dec.Covevent.covev_loc) ->
          begin
            (t.model).mod_conds <- 
              cond :: (t.model).mod_conds;
            populate_vector_model cond dec l;
          end
      | t :: l -> populate_vector_model cond dec l
    in
    let rec populate_mcdc cond = function
	[] -> ()
      | dec :: l -> 
          populate_vector_model cond dec mcdc_list;
          populate_mcdc cond l
    in
    let rec second_pass = function
	[] -> ()
      | cond :: l ->
          begin 
            match cond.Covevent.covev_status with 
		Covevent.Coverage_event_condition dec_list ->
                  begin 
                    populate_mcdc cond dec_list; 
                    second_pass l 
                  end
              | _ -> assert false
          end
    in 
      second_pass cond_list;
       (mcdc_list ,cond_list, dec_list)
      
