(***********************************************************************)
(*                              Project Couverture                     *)
(*                                                                     *)
(* file: covutils.ml                                                   *)
(* authors: Adrien Jonquet, Mathias Bourgoin                           *)
(* licence: CeCILL-B                                                   *)
(***********************************************************************)

open Misc
open Asttypes
open Primitive
open Path
open Types
open Typedtree
open Typeopt
open Lambda

type from_dec = 
   False 
 | True of lambda_covevent list

let string_of_expression_desc = function
    Texp_ident (path, {val_kind = Val_prim p}) -> "Texp_ident Val_prim"
  | Texp_ident(path, {val_kind = Val_anc _}) -> "Texp_ident Val_anc"
  | Texp_ident(path, {val_kind = Val_reg}) -> "Texp_ident Val_reg"
  | Texp_ident(path, {val_kind = Val_self _}) -> "Texp_ident Val_self"
  | Texp_ident _ ->  "Texp_ident failwith"
  | Texp_constant _ -> "Texp_constant"
  | Texp_let (_, _, _) -> "Texp_let"
  | Texp_function (_, _) -> "Texp_function" 
  | Texp_apply ({exp_desc = Texp_ident(path, {val_kind = Val_prim p})}, oargs) -> "Texp_apply prim"
  | Texp_apply (_, _) -> "Texp_apply"
  | Texp_match (_, _, _) -> "Texp_match"
  | Texp_try (_, _) -> "Texp_try"
  | Texp_tuple _ -> "Texp_tuple"
  | Texp_construct (cstr, _) -> 
      (match cstr.cstr_tag with 
	   Cstr_constant n -> "Texp_construct constant"
	 | Cstr_block n -> "Texp_construct block"
	 | Cstr_exception path -> "Texp_construct ")
  | Texp_variant (_, _) -> "Texp_variant"
  | Texp_record (_, _) -> "Texp_record"
  | Texp_field (_, _) -> "Texp_field"
  | Texp_setfield (_, _, _) -> "Texp_setfield"
  | Texp_array _ -> "Texp_array"
  | Texp_ifthenelse (_, _, _) -> "Texp_ifthenelse"
  | Texp_sequence (_, _) -> "Texp_sequence"
  | Texp_while (_, _) -> "Texp_while"
  | Texp_for (_, _, _, _, _) -> "Texp_for"
  | Texp_when (_, _) -> "Texp_when"
  | Texp_send (_, _) -> "Texp_send"
  | Texp_new (_, _) -> "Texp_new"
  | Texp_instvar (_, _) -> "Texp_instvar"
  | Texp_setinstvar (_, _, _) -> "Texp_setinstvar"
  | Texp_override (_, _) -> "Texp_override"
  | Texp_letmodule (_, _, _) -> "Texp_letmodule"
  | Texp_assert _ -> "Texp_assert"
  | Texp_assertfalse -> "Texp_assertfalse"
  | Texp_lazy _ -> "Texp_lazy"
  | Texp_object (_, _, _) -> "Texp_object"
	
let print s = 
  (try 
     if (Sys.getenv "DEBUG") =  "1" then
       print_string s
   with
     | _ -> ())
     
let rec is_bool = function
    Tconstr ((Pident i) as path, _, _) -> 
      begin 
	print "   Tconstr ";
	begin
	  if path = Predef.path_bool then (print "bool\n" ; true)
	  else (print ((Ident.name i)^"\n"); false)
	end
      end
  | Tconstr (_, _, _) -> false
  | Tpoly (exp, exp_l) -> 
      begin 
	  print "   Tpoly\n";
	  false
	end
    | Tvariant _ -> 
	begin 
	  print "   Tvariant\n";
	  false
	end
    | Tsubst te -> 
	begin
	  print "   Tsubst\n";
	  is_bool te.desc 
	end
    | Tlink te -> 
	begin
	  print "   Tlink\n";
	  is_bool te.desc
	end
    | Tfield (_, _, _, _) -> 
	begin
	  print "   Tfield\n";
	  false
	end
    | Tobject (_, _) -> 
	begin 
	  print "   Tobject\n";
	  false
	end
    | Ttuple _ -> 
	begin
	  print "   Ttuple\n";
	  false
	end
    | Tarrow (_, _, _, _) -> 
	begin 
	  print "   Tarrow\n";
	  false
	end
    | Tunivar -> 
	begin 
	  print "   Tunivar\n";
	  false
	end
    | Tnil  -> 
	begin 
	  print "   Tnil\n";
	  false
	end
    | Tvar ->
	begin 
	  print "   Tvar\n";
	  false
	end
     
(* Insertion of coverage events *)
let put_before exp lam status =
  Lcovevent(lam, { lcovev_loc = exp.exp_loc;
		   lcovev_kind = Lcovev_before;
		   lcovev_status = status })
    
let put_after exp lam status =
  Lcovevent(lam, { lcovev_loc = exp.exp_loc;
		   lcovev_kind = Lcovev_after exp.exp_type;
		   lcovev_status = status })
    
    
let put_covev exp lam status =
    put_after exp (put_before exp lam status) status

let build_lambda_covev exp kind status =
  { lcovev_loc = exp.exp_loc;
    lcovev_kind = 
      if kind = "before" then
        Lcovev_before
      else   
        Lcovev_after exp.exp_type;
    lcovev_status = status }
    

